package main

import (
	"slices"
	"testing"
)

var reTests = []struct {
	re     string
	str    string
	result []matchIndex
}{
	{"a", "abc", []matchIndex{{0, 1}}},
	{"a", "bca", []matchIndex{{2, 3}}},
	{"l", "ggllgg", []matchIndex{{2, 3}, {3, 4}}},
	{"(b|c)", "abdceb", []matchIndex{{1, 2}, {3, 4}, {5, 6}}},
	{"a+", "brerereraaaaabbbbb", []matchIndex{{8, 13}}},
	{"ab+", "qweqweqweaqweqweabbbbbr", []matchIndex{{16, 22}}},
	{"(b|c|A)", "ooaoobocA", []matchIndex{{5, 6}, {7, 8}, {8, 9}}},
	{"ab*", "a", []matchIndex{{0, 1}}},
	{"ab*", "abb", []matchIndex{{0, 3}}},
	{"a*b", "aaab", []matchIndex{{0, 4}}},
	{"a*b", "qwqw", []matchIndex{}},
	{"(abc)*", "abcabcabc", []matchIndex{{0, 9}, {9, 9}}},
	{"((abc)|(def))*", "abcdef", []matchIndex{{0, 6}, {6, 6}}},
	{"(abc)*|(def)*", "abcdef", []matchIndex{{0, 3}, {3, 6}, {6, 6}}},
	{"b*a*a", "bba", []matchIndex{{0, 3}}},
	{"(ab)+", "abcabddd", []matchIndex{{0, 2}, {3, 5}}},
	{"a(b(c|d)*)*", "abccbd", []matchIndex{{0, 6}}},
	{"a(b|c)*d+", "abccdd", []matchIndex{{0, 6}}},
	{"a*", "", []matchIndex{{0, 0}}},
	{"a|b", "c", []matchIndex{}},
	{"(a|b)*c", "aabbc", []matchIndex{{0, 5}}},
	{"a(b|b)", "ab", []matchIndex{{0, 2}}},
	{"a*", "aaaaaaaa", []matchIndex{{0, 8}, {8, 8}}},

	{"ab?", "ab", []matchIndex{{0, 2}}},
	{"a?b", "ab", []matchIndex{{0, 2}}},
	{"a?", "", []matchIndex{{0, 0}}},
	{"a?b?c", "a", []matchIndex{}},
	{"a?b?c?", "ab", []matchIndex{{0, 2}, {2, 2}}},
	{"a?b?c?", "ac", []matchIndex{{0, 2}, {2, 2}}},
	{"a?b?c", "abc", []matchIndex{{0, 3}}},
	{"a?b?c", "acb", []matchIndex{{0, 2}}},

	{"[abc]", "defadefbdefce", []matchIndex{{3, 4}, {7, 8}, {11, 12}}},
	{"[ab]c", "ab", []matchIndex{}},
	{"g[ab]c", "gac", []matchIndex{{0, 3}}},
	{"g[ab]c", "gbc", []matchIndex{{0, 3}}},
	{"g[ab]c", "gc", []matchIndex{}},
	{"g[ab]c", "gfc", []matchIndex{}},
	{"[ab]*", "aabbbabaababab", []matchIndex{{0, 14}, {14, 14}}},
	{"[ab]+", "aabbbablaababab", []matchIndex{{0, 7}, {8, 15}}},
	{"[Ff]r[Uu]it", "fruit", []matchIndex{{0, 5}}},
	{"[Ff]r[Uu]it", "FrUit", []matchIndex{{0, 5}}},
	{"[Ff]r[Uu|]it", "Fr|it", []matchIndex{{0, 5}}},
	{"[Ff]r([Uu]|[pP])it", "Frpit", []matchIndex{{0, 5}}},
	{"[Ff]r[Uu]|[pP]it", "Frpit", []matchIndex{{2, 5}}},
	{"[a-zA-Z]+", "Hello, how is it going?", []matchIndex{{0, 5}, {7, 10}, {11, 13}, {14, 16}, {17, 22}}},

	{".+", "Hello, how is it going?", []matchIndex{{0, 23}}},
	{"a.", "a ", []matchIndex{{0, 2}}},
	{"a.b", "a/b", []matchIndex{{0, 3}}},
	{".", "a ", []matchIndex{{0, 1}, {1, 2}}},
	{"a.", "a ", []matchIndex{{0, 2}}},
	{".+b", "abc", []matchIndex{{0, 2}}},

	{`\d`, "1a0a3s'''34343s", []matchIndex{{0, 1}, {2, 3}, {4, 5}, {9, 10}, {10, 11}, {11, 12}, {12, 13}, {13, 14}}},
	{`\\`, `a\b\c\qwe\`, []matchIndex{{1, 2}, {3, 4}, {5, 6}, {9, 10}}},
	{`\W`, `"Hello", he said. How are you doing?`, []matchIndex{{0, 1}, {6, 7}, {7, 8}, {8, 9}, {11, 12}, {16, 17}, {17, 18}, {21, 22}, {25, 26}, {29, 30}, {35, 36}}},
	{`\w`, ";';';';';'qwe12", []matchIndex{{10, 11}, {11, 12}, {12, 13}, {13, 14}, {14, 15}}},
	{`\s`, "a b c 	d", []matchIndex{{1, 2}, {3, 4}, {5, 6}, {6, 7}}},
	{`\<`, "<HTML><body>", []matchIndex{{0, 1}, {6, 7}}},
	{`\(.+\)`, "Not (paranthesized), (so) is (this) not", []matchIndex{{4, 35}}},

	{"[^abc]+", "qarbtopsaplpclkpasdmb prejip0r,p", []matchIndex{{0, 1}, {2, 3}, {4, 8}, {9, 12}, {13, 16}, {17, 20}, {21, 32}}},
	{"[^a]+", "qqqaq", []matchIndex{{0, 3}, {4, 5}}},
	{"[^0-9]+", "a1b2c3dd", []matchIndex{{0, 1}, {2, 3}, {4, 5}, {6, 8}}},
	{"[^abc]+", "ababababbababaccacacacaca", []matchIndex{}},
	{`\([^)]+\)`, "Not (paranthesized), (so) is (this) not", []matchIndex{{4, 19}, {21, 25}, {29, 35}}},

	{"^ab", "ab bab", []matchIndex{{0, 2}}},
	{"^aaaa^", "aaaaaaaa", []matchIndex{}},
	{"^([bB][Gg])", "bG", []matchIndex{{0, 2}}},
	{"b$", "ba", []matchIndex{}},
	{"(boy|girl)$", "girlf", []matchIndex{}},
	{`\bint\b`, "print int integer", []matchIndex{{6, 9}}},
	{`int\b`, "ints", []matchIndex{}},
	{`int(\b|a)`, "inta", []matchIndex{{0, 4}}},
	{`\b\d+\b`, "511 a3 43", []matchIndex{{0, 3}, {7, 9}}},
	{`\Bint\B`, "prints int integer print", []matchIndex{{2, 5}}},
	{`^`, "5^3^2", []matchIndex{{0, 0}}},
	{`\^`, "5^3^2", []matchIndex{{1, 2}, {3, 4}}},
	{`pool$`, "pool carpool", []matchIndex{{8, 12}}},
	{`^int$`, "print int integer", []matchIndex{}},
	{`^int$`, "int", []matchIndex{{0, 3}}},
	{`b*`, "aaaaaaaaaaqweqwe", []matchIndex{{0, 0}, {1, 1}, {2, 2}, {3, 3}, {4, 4}, {5, 5}, {6, 6}, {7, 7}, {8, 8}, {9, 9}, {10, 10}, {11, 11}, {12, 12}, {13, 13}, {14, 14}, {15, 15}, {16, 16}}},

	{"a{4}", "aabaaa", []matchIndex{}},
	{"ab{5}", "abbbbbab", []matchIndex{{0, 6}}},
	{"(a|b){3,4}", "aba", []matchIndex{{0, 3}}},
	{"(a|b){3,4}", "ababaa", []matchIndex{{0, 4}}},
	{"(bc){5,}", "bcbcbcbcbcbcbcbc", []matchIndex{{0, 16}}},
	{`\d{3,4}`, "1209", []matchIndex{{0, 4}}},
	{`\d{3,4}`, "109", []matchIndex{{0, 3}}},
	{`\d{3,4}`, "5", []matchIndex{}},
	{`\d{3,4}`, "123135", []matchIndex{{0, 4}}},
	{`\d{3,4}`, "89a-0", []matchIndex{}},
	{`\d{3,4}`, "ababab555", []matchIndex{{6, 9}}},
	{`\bpaint\b`, "paints", []matchIndex{}},
	{`\b\w{5}\b`, "paint", []matchIndex{{0, 5}}},
}

func TestFindAllMatches(t *testing.T) {
	for _, test := range reTests {
		t.Run(test.re+"	"+test.str, func(t *testing.T) {
			re_postfix := shuntingYard(test.re)
			startState := thompson(re_postfix)
			matchIndices := findAllMatches(startState, test.str)
			if !slices.Equal(test.result, matchIndices) {
				t.Errorf("Wanted %v	Got %v\n", test.result, matchIndices)
			}
		})
	}
}
